//
//  ONOpenCLArchive.m
//  clc
//
//  Created by Andrew Finnell on 12/14/09.
//  Copyright 2011 Fortunate Bear, LLC. All rights reserved.
//

#import "ONOpenCLArchive.h"
#import <zlib.h>

NSString * const ONOpenCLArchiveExtension = @"clar";
NSString * const ONOpenCLArchiveErrorDomain = @"ONOpenCLArchiveErrorDomain";

static const NSInteger ONOpenCLArchiveErrorCompression = 1000;
static const NSInteger ONOpenCLArchiveErrorMissingKeys = 1001;
static const NSInteger ONOpenCLArchiveErrorUnknownCompression = 1002;

static const NSInteger ONOpenCLArchiveCurrentVersion = 1;
static NSString * const ONOpenCLArchiveCompressionZlib = @"Zlib";

@implementation ONOpenCLArchive

+ (NSData *) dataWithOpenCLSource:(NSString *)source error:(NSError **)error
{
	const char *rawSource = [source UTF8String];
	NSUInteger rawSourceLength = strlen(rawSource);
	
	NSUInteger compressedBufferSize = (NSUInteger)ceilf(rawSourceLength * 1.10) + 12;
	NSMutableData *compressedBuffer = [NSMutableData dataWithLength:compressedBufferSize];
	uLongf finalSize = compressedBufferSize;
	int result = compress([compressedBuffer mutableBytes], &finalSize, (const unsigned char *)rawSource, rawSourceLength);
	
	if ( result != Z_OK ) {
		*error = [NSError errorWithDomain:ONOpenCLArchiveErrorDomain 
									 code:ONOpenCLArchiveErrorCompression 
								 userInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:result] forKey:@"code"]];
		return nil;
	}
	
	[compressedBuffer setLength:finalSize];
	
	NSDictionary *plist = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInteger:ONOpenCLArchiveCurrentVersion], @"Version", 
						   ONOpenCLArchiveCompressionZlib, @"Compression",
						   compressedBuffer, @"Data",
						   [NSNumber numberWithUnsignedInteger:rawSourceLength], @"UncompressedSize",
						   nil];
	return [NSPropertyListSerialization dataWithPropertyList:plist format:NSPropertyListBinaryFormat_v1_0 options:0 error:error];
}

+ (NSString *) openCLSourceWithData:(NSData *)data error:(NSError **)error
{
	NSPropertyListFormat format = 0;
	NSDictionary *plist = [NSPropertyListSerialization propertyListWithData:data options:NSPropertyListImmutable format:&format error:error];
	
	if ( plist == nil || ![plist isKindOfClass:[NSDictionary class]] )
		return nil;

	NSString *compression = [plist objectForKey:@"Compression"];
	NSNumber *uncompressedSize = [plist objectForKey:@"UncompressedSize"];
	NSData *compressedData = [plist objectForKey:@"Data"];
	
	if ( compression == nil || uncompressedSize == nil || compressedData == nil ) {
		*error = [NSError errorWithDomain:ONOpenCLArchiveErrorDomain 
									 code:ONOpenCLArchiveErrorMissingKeys 
								 userInfo:[NSDictionary dictionaryWithObject:@"Missing keys from archive" forKey:@"reason"]];
		
		return nil;
	}
	
	if ( ![compression isEqualToString:ONOpenCLArchiveCompressionZlib] ) {
		*error = [NSError errorWithDomain:ONOpenCLArchiveErrorDomain 
									 code:ONOpenCLArchiveErrorUnknownCompression
								 userInfo:[NSDictionary dictionaryWithObject:@"Unknown compression scheme" forKey:@"reason"]];
		
		return nil;
	}
	
	NSMutableData *uncompressedBuffer = [NSMutableData dataWithLength:[uncompressedSize unsignedIntegerValue]];
	uLongf finalSize = [uncompressedSize unsignedIntegerValue];
	int result = uncompress([uncompressedBuffer mutableBytes], &finalSize, [compressedData bytes], [compressedData length]);
	
	if ( result != Z_OK ) {
		*error = [NSError errorWithDomain:ONOpenCLArchiveErrorDomain 
									 code:ONOpenCLArchiveErrorCompression 
								 userInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:result] forKey:@"code"]];
		return nil;
	}

	return [[[NSString alloc] initWithData:uncompressedBuffer encoding:NSUTF8StringEncoding] autorelease];
}

+ (NSString *) openCLSourceWithFile:(NSString *)file error:(NSError **)error
{
	NSData *data = [NSData dataWithContentsOfFile:file];

	return [self openCLSourceWithData:data error:error];
}

@end

@implementation NSBundle(NSBundleOpenCLArchiveExtension)

- (NSString *)pathForOpenCLResource:(NSString *)name
{
	return [self pathForResource:name ofType:ONOpenCLArchiveExtension];
}

- (NSArray *)pathsForOpenCLResources
{
	return [self pathsForResourcesOfType:ONOpenCLArchiveExtension inDirectory:nil];
}

- (NSString *)openCLResource:(NSString *)name
{
	NSString *path = [self pathForOpenCLResource:name];
	if ( path == nil )
		return nil;
	
	NSData *data = [NSData dataWithContentsOfFile:path];
	NSError *error = nil;
	return [ONOpenCLArchive openCLSourceWithData:data error:&error];
}

@end
