#import <Foundation/Foundation.h>
#import <OpenCL/OpenCL.h>
#import <unistd.h>
#import "ONOpenCLArchive.h"

static const char * DeviceTypeToString(cl_device_type deviceType)
{
	switch (deviceType) {
		case CL_DEVICE_TYPE_CPU:
			return "cpu";
		case CL_DEVICE_TYPE_GPU:
			return "gpu";
		case CL_DEVICE_TYPE_ACCELERATOR:
			return "accelerator";
	}
	
	return "(nil)";
}

static NSString *ResolvePath(NSString *path)
{
	// If it's already an absolute path, we're done
	path = [path stringByStandardizingPath];
	if ( [path isAbsolutePath] )
		return path;
	
	// Assume it's a relative path
	char *directory = getcwd(nil, 0);
	NSString *currentDirectory = [NSString stringWithUTF8String:directory];
	free(directory);
	
	return [[currentDirectory stringByAppendingPathComponent:path] stringByStandardizingPath];
}

static void ReportErrors(NSString *allErrors, NSString *fileName)
{
	NSArray *lines = [allErrors componentsSeparatedByCharactersInSet:[NSCharacterSet newlineCharacterSet]];
	for (NSString *line in lines) {
		if ( [line hasPrefix:@":"] )
			fprintf(stderr, "%s%s\n", [fileName UTF8String], [line UTF8String]);
		else if ( [line hasPrefix:@"<program source>:"] ) {
			NSString *errorString = [line substringFromIndex:[@"<program source>" length]];
			fprintf(stderr, "%s%s\n", [fileName UTF8String], [errorString UTF8String]);			
		} else if ( ![line hasPrefix:@"\t"] && ![line hasPrefix:@" "] && ![line hasPrefix:@"\n"] && [line length] > 0 ) {
			fprintf(stderr, "%s:1:1: error: %s\n", [fileName UTF8String], [line UTF8String]);
		}
	}
}

static NSString *DefaultOutputFile(NSString *userSpecifiedFile, NSString *inputFile)
{
	if ( userSpecifiedFile != nil )
		return userSpecifiedFile;
	
	NSString *fileName = [[[inputFile lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:ONOpenCLArchiveExtension];
	return ResolvePath(fileName);
}

static void WriteArchive(NSString *fileName, NSString *source)
{
	NSError *error = nil;
	NSData *data = [ONOpenCLArchive dataWithOpenCLSource:source error:&error];
	[data writeToFile:fileName atomically:NO];
}

static BOOL ProcessFile(NSString *fileName, cl_device_type deviceType, NSString *outputFile) {
	// Grab the type of device they want to compile against
	cl_device_id device = nil;
	int err = clGetDeviceIDs(nil, deviceType, 1, &device, nil);
	if (err != CL_SUCCESS) {
		fprintf(stderr, "Unable to find device of type %s. err = %d\n", DeviceTypeToString(deviceType), err);
		return YES;
	}
	
	// Create a context to operate in
	cl_context context = clCreateContext(0, 1, &device, nil, nil, &err);
	if ( context == nil ) {
		fprintf(stderr, "Unable to create context. err = %d\n", err);
		return YES;
	}
	
	// Read in the file
	NSError *error = nil;
	NSString *source = [NSString stringWithContentsOfFile:fileName encoding:NSUTF8StringEncoding error:&error];
	const char *sourceString = [source UTF8String];
	cl_program program = clCreateProgramWithSource(context, 1, &sourceString, nil, &err);
	if ( program == nil || err != CL_SUCCESS) {
		clReleaseContext(context);
		fprintf(stderr, "Unable to create program. err = %d\n", err);
		return YES;
	}
	
	// Build the program executable
	err = clBuildProgram(program, 0, nil, "-Werror", nil, nil);
	if (err != CL_SUCCESS) {
		// If we got errors report them. Silence means success
		size_t length = 0;			
		clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, 0, nil, &length);
		
		char *buffer = calloc(length + 1, sizeof(char));
		clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, length, buffer, &length);		
		ReportErrors([NSString stringWithUTF8String:buffer], fileName);
		free(buffer);
	} else 
		WriteArchive(DefaultOutputFile(outputFile, fileName), source);

	clReleaseProgram(program);
	clReleaseContext(context);

	
	return err != CL_SUCCESS;
}


static void Usage(const char *name)
{
	fprintf(stdout, "usage: %s [-d cpu|gpu|accelarator] [-o outputFile] files\n", name);
}

int main (int argc, const char * argv[]) {
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];

	// Process any parameters
	cl_device_type deviceType = CL_DEVICE_TYPE_CPU;
	NSMutableArray *files = [NSMutableArray arrayWithCapacity:argc];
	NSString *outputFile = nil;
	for (int i = 1; i < argc; ++i) {
		NSString *param = [NSString stringWithUTF8String:argv[i]];
		if ( [param isEqualToString:@"-d"] ) {
			if ( i != (argc - 1) ) {
				NSString *typeString = [NSString stringWithUTF8String:argv[++i]];
				if ( [typeString isEqualToString:@"cpu"] )
					deviceType = CL_DEVICE_TYPE_CPU;
				else if ( [typeString isEqualToString:@"gpu"] )
					deviceType = CL_DEVICE_TYPE_GPU;
				else if ( [typeString isEqualToString:@"accelerator"] )
					deviceType = CL_DEVICE_TYPE_ACCELERATOR;
				else {
					Usage(argv[0]);
					return 1;
				}
			} else {
				Usage(argv[0]);
				return 1;
			}
		} else if ( [param isEqualToString:@"-o"] ) {
			if ( i != (argc - 1) ) {
				outputFile = [NSString stringWithUTF8String:argv[++i]];
			} else {
				Usage(argv[0]);
				return 1;
			}			
		} else {
			// Assume this is a file name
			[files addObject:ResolvePath(param)];
		}
	}
	
	// Process any files
	BOOL hasErrors = NO;
	for (NSString *filePath in files)
		hasErrors |= ProcessFile(filePath, deviceType, outputFile);
	
    [pool drain];
	
	if ( hasErrors )
		return 1;
	
    return 0;
}
